#include <stdlib.h>
#include <time.h>

#include "solid.hpp"

// ================================================================
//  Class Solid
//
//  Solid is an implementation of the game "Solitaire Dice,"
//  invented by Sid Sackson ["A Gamut of Games", Random House 1969]
//
//  This module implements the pure game logic with no assumptions
//  about the user interface.
//  Only minimal validation of user input is implemented,
//  since this responsibility really belongs to the user interface.
// ================================================================

//  Ctor

Solid::Solid( int seed )
{
    int k;

    // Initialize RNG if seed is nonzero

    if (seed)
        srand( (unsigned int) time(NULL) );
        
    // Initialize state variables

    numRejects = 0;
    freebie = 0;

    for (k=0; k < 13; k++)
    {
        hits[k] = 0;
    }

    for (k=0; k < 5; k++)
    {
        die[k] = 0;
        group[k] = -1;
    }
    
    for (k=1; k <= 6; k++)
        rejectCount[k] = 0;
}

//  Inspectors

int
Solid::Die( int n ) const
{
    if (n < 0 || n > 4)
        return 0;

    return die[n];
}

int
Solid::Group( int n ) const
{
    if (n < 0 || n > 4)
        return -1;

    return group[n];
}

int
Solid::Hits( int n ) const
{
    if ( n < 2 || n > 12 )
        return -1;

    return hits[n];
}

int
Solid::NumRejects() const
{
    return numRejects;
}

int
Solid::RejectCounts( int n ) const
{
    if (n < 1 || n > 6)
        return -1;

    return rejectCount[n];
}

static int
clip( int n )
{
    if (n < 5)
        return 0;
    else if (n > 10)
        return 5;
    else
        return n-5;
}

int
Solid::Score() const
{
    int score;

    int weight2 = 100;
    int weight3 = 70;
    int weight4 = 60;
    int weight5 = 50;
    int weight6 = 40;
    int weight7 = 30;

    // Add up positive scores

    score = weight2 * ( clip( hits[2] ) + clip( hits[12] ) ) +
            weight3 * ( clip( hits[3] ) + clip( hits[11] ) ) +
            weight4 * ( clip( hits[4] ) + clip( hits[10] ) ) +
            weight5 * ( clip( hits[5] ) + clip( hits[9] ) ) +
            weight6 * ( clip( hits[6] ) + clip( hits[8] ) ) +
            weight7 * clip( hits[7] );

    // Levy penalties

    for (int k=2; k < 13; k++)
    {
        if ( 0 < hits[k] && hits[k] < 5 )
            score -= 200;
    }
    
    return score;
}

int
Solid::Over() const
{
    int k;

    for ( k=1; k <= 6; k++ )
        if ( rejectCount[k] >= 8 )
            return 1;
    return 0;
}

//  Mutators

int
Solid::CommitTurn()
{
    int h1 = 0;
    int h2 = 0;

    int k;
    
    //  Calculate pair values

    for ( k=0; k < 5; k++ )
    {
        switch( group[k] )
        {
            case 1:
                h1 += die[k];
                break;

            case 2:
                h2 += die[k];
                break;
        }
    }

    // Sanity checks

    if ( !h1 || !h2 )
        return 2;

    hits[ h1 ]++;
    hits[ h2 ]++;

    return 0;
}


// Assign die d to group g

int
Solid::SetGroup( int d, int g )
{
    if (d < 0 || d > 4)
        return 1;

    if (g < 0 || g > 2)
        return 2;
    
    if ( g )
    {
        group[d] = g;
    }
    else if (!freebie)
    {
        // Handle reject

        int pips;   // the number shown on die[d]

        group[d] = 0;

        pips = die[d];

        // Is it one of the existing rejects?
        
        if ( rejectCount[ pips ] )
        {
            rejectCount[ pips ]++;
        }
        else if (numRejects == 3)
        {
            // Whoa! Can't have more than 3 rejects
            return 3;
        }
        else
        {
            // new reject
            rejectCount[ pips ] = 1;
            numRejects++;
        }
    }

    return 0;
}

int roll()
{
    int n;
    n = rand() >> 6;

    return 1 + (n % 6);
}


// Note the logic: if there are less than 3 reject numbers, then
// this can't be a freebie.  If there ARE 3 reject numbers, then
// assume it IS a freebie until proven otherwise.

int
Solid::Roll()
{
    int k;

    if ( numRejects < 3 )
        freebie = 0;
    else
        freebie = 1;

    for ( k=0; k < 5; k++ )
    {
        group[k] = -1;

        die[k] = roll();

        if ( numRejects == 3 && rejectCount[ die[k] ] > 0 )
            freebie = 0;
    }

    return 0;
}
